# -*- coding: utf-8 -*-
'''
    pygeodiff.main
    --------------
    Main entry of the library
    :copyright: (c) 2019 Peter Petrik
    :license: MIT, see LICENSE for more details.
'''

from .geodifflib import GeoDiffLib


class GeoDiff:
    """
        geodiff is a module to create and apply changesets to GIS files (geopackage)
    """
    
    def __init__(self, libname=None):
        """
            if libname is None, it tries to import c-extension from wheel
            messages are shown in stdout/stderr. 
            Use environment variable GEODIFF_LOGGER_LEVEL 0(Nothing)-4(Debug) to
            set level (Errors by default)
        """
        self.clib = GeoDiffLib(libname)

    def set_logger_callback(self, callback):
        """
            Assign custom logger
            Replace default stdout/stderr logger with custom.
        """
        return self.clib.set_logger_callback(callback)

    LevelError = 1
    LevelWarning = 2
    LevelInfo = 3
    LevelDebug = 4

    def set_maximum_logger_level(self, maxLevel):
        """
           Assign maximum level of messages that are passed to logger callbac
           Based on maxLogLevel, the messages are filtered by level:
           maxLogLevel = 0 nothing is passed to logger callback
           maxLogLevel = 1 errors are passed to logger callback
           maxLogLevel = 2 errors and warnings are passed to logger callback
           maxLogLevel = 3 errors, warnings and infos are passed to logger callback
           maxLogLevel = 4 errors, warnings, infos, debug messages are passed to logger callback
        """
        return self.clib.set_maximum_logger_level(maxLevel)

    def create_changeset(self, base, modified, changeset):
        """
            Creates changeset file (binary) in such way that
            if CHANGESET is applied to BASE by applyChangeset,
            MODIFIED will be created

            BASE --- CHANGESET ---> MODIFIED

            \param base [input] BASE sqlite3/geopackage file
            \param modified [input] MODIFIED sqlite3/geopackage file
            \param changeset [output] changeset between BASE -> MODIFIED

            raises SqliteDiffError on error
        """
        return self.clib.create_changeset(base, modified, changeset)

    def invert_changeset(self, changeset, changeset_inv):
        """
            Inverts changeset file (binary) in such way that
            if CHANGESET_INV is applied to MODIFIED by applyChangeset,
            BASE will be created

            \param changeset [input] changeset between BASE -> MODIFIED
            \param changeset_inv [output] changeset between MODIFIED -> BASE

            \returns number of conflics

            raises SqliteDiffError on error
        """
        return self.clib.invert_changeset(changeset, changeset_inv)

    def rebase(self, base, modified_their, modified, conflict):
        """
            Rebases local modified version from base to modified_their version

                   --- > MODIFIED_THEIR
            BASE -|
                   ----> MODIFIED (local) ---> MODIFIED_THEIR_PLUS_MINE

            Steps performed on MODIFIED (local) file:
            1. undo local changes MODIFIED -> BASE
            2. apply changes from MODIFIED_THEIR
            3. apply rebased local changes and create MODIFIED_THEIR_PLUS_MINE

            Note, when rebase is not successfull, modified could be in random state.
            This works in general, even when base==modified, or base==modified_theirs

             \param base [input] BASE sqlite3/geopackage file
             \param modified_their [input] MODIFIED sqlite3/geopackage file
             \param modified [input] local copy of the changes to be rebased
             \param conflict [output] file where the automatically resolved conflicts are stored. If there are no conflicts, file is not created

             raises SqliteDiffError on error
        """
        return self.clib.rebase(base, modified_their, modified, conflict)

    def create_rebased_changeset(self, base, modified, changeset_their, changeset, conflict):
        """
            Creates changeset file (binary) in such way that
            if CHANGESET is applied to MODIFIED_THEIR by
            applyChangeset, the new state will contain all
            changes from MODIFIED and MODIFIED_THEIR.

                   --- CHANGESET_THEIR ---> MODIFIED_THEIR --- CHANGESET ---> MODIFIED_THEIR_PLUS_MINE
            BASE -|
                   -----------------------> MODIFIED

             \param base [input] BASE sqlite3/geopackage file
             \param modified [input] MODIFIED sqlite3/geopackage file
             \param changeset_their [input] changeset between BASE -> MODIFIED_THEIR
             \param changeset [output] changeset between MODIFIED_THEIR -> MODIFIED_THEIR_PLUS_MINE
             \param conflict [output] file where the automatically resolved conflicts are stored. If there are no conflicts, file is not created

             raises SqliteDiffError on error
        """
        return self.clib.create_rebased_changeset(base, modified, changeset_their, changeset, conflict)

    def apply_changeset(self, base, changeset):
        """
            Applies changeset file (binary) to BASE

            \param base [input/output] BASE sqlite3/geopackage file
            \param changeset [input] changeset to apply to BASE
            \returns number of conflicts

            raises SqliteDiffError on error
        """
        return self.clib.apply_changeset(base, changeset)

    def list_changes(self, changeset, json):
        """
            Lists changeset content JSON file
            JSON contains all changes in human/machine readable name
            \returns number of changes

             raises SqliteDiffError on error
        """
        return self.clib.list_changes(changeset, json)

    def list_changes_summary(self, changeset, json):
        """
            Lists changeset summary content JSON file
            JSON contains a list of how many inserts/edits/deletes is contained in changeset for each table
            \returns number of changes

             raises SqliteDiffError on error
        """
        return self.clib.list_changes_summary(changeset, json)


    def has_changes(self, changeset):
        """
            \returns whether changeset contains at least one change

            raises SqliteDiffError on error
        """
        return self.clib.has_changes(changeset)

    def changes_count(self, changeset):
        """
            \returns number of changes

             raises SqliteDiffError on error
        """
        return self.clib.changes_count(changeset)

    def version(self):
        """
            geodiff version
        """
        return self.clib.version()


def main():
    diff_lib = GeoDiff()
    print("pygeodiff " + diff_lib.version())



# -*- coding: utf-8 -*-
'''
    pygeodiff.geodifflib
    --------------------
    This module provides wrapper of geodiff C library
    :copyright: (c) 2019 Peter Petrik
    :license: MIT, see LICENSE for more details.
'''

import ctypes
import os
import platform
from ctypes.util import find_library
from .__about__ import __version__


class GeoDiffLibError(Exception):
    pass


class GeoDiffLibConflictError(GeoDiffLibError):
    pass


class GeoDiffLibVersionError(GeoDiffLibError):
    pass


# keep in sync with c-library
SUCCESS=0
ERROR=1
CONFLICT=2

def _parse_return_code(rc, msg):
    if rc == SUCCESS:
        return
    elif rc == ERROR:
        raise GeoDiffLibError(msg)
    elif rc == CONFLICT:
        raise GeoDiffLibConflictError(msg)

class GeoDiffLib:
    def __init__(self, name):
        if name is None:
            self.libname = self.package_libname()
            if not os.path.exists(self.libname):
                # not found, try system library
                self.libname = find_library("geodiff")
        else:
            self.libname = name

        if self.libname is None:
            raise GeoDiffLibVersionError("Unable to locate GeoDiff library, tried " + self.package_libname() + " and geodiff on system.")

        self.lib = ctypes.CDLL(self.libname, use_errno=True)
        self.init()
        self.check_version()

    def package_libname(self):
        # assume that the package is installed through PIP
        if platform.system() == 'Windows':
            prefix = ""
            arch = platform.architecture()[0]  # 64bit or 32bit
            if "32" in arch:
                suffix = "-win32.dll"
            else:
                suffix = ".dll"
        elif platform.system() == 'Darwin':
            prefix = "lib"
            suffix = ".dylib"
        else:
            prefix = "lib"
            suffix = ".so"
        whl_lib = prefix + 'pygeodiff-' + __version__ + '-python' + suffix
        dir_path = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dir_path, whl_lib)

    def init(self):
        func = self.lib.GEODIFF_init
        func()

    def set_logger_callback(self, callback):
        func = self.lib.GEODIFF_setLoggerCallback
        CMPFUNC = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.c_char_p)
        func.argtypes = [CMPFUNC]
        # do not remove self, callback needs to be member
        self.callbackLogger = CMPFUNC(callback)
        func(self.callbackLogger)

    def set_maximum_logger_level(self, maxLevel):
        func = self.lib.GEODIFF_setMaximumLoggerLevel
        func.argtypes = [ctypes.c_int]
        func(maxLevel)

    def version(self):
        func = self.lib.GEODIFF_version
        func.restype = ctypes.c_char_p
        ver = func()
        return ver.decode('utf-8')

    def check_version(self):
        cversion = self.version()
        pyversion = __version__
        if cversion != pyversion:
            raise GeoDiffLibVersionError("version mismatch (%s C vs %s PY)".format(cversion, pyversion))

    def create_changeset(self, base, modified, changeset):
        func = self.lib.GEODIFF_createChangeset
        func.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = base.encode('utf-8')
        b_string2 = modified.encode('utf-8')
        b_string3 = changeset.encode('utf-8')

        res = func(b_string1, b_string2, b_string3)
        _parse_return_code(res, "createChangeset")

    def invert_changeset(self, changeset, changeset_inv):
        func = self.lib.GEODIFF_invertChangeset
        func.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = changeset.encode('utf-8')
        b_string2 = changeset_inv.encode('utf-8')

        res = func(b_string1, b_string2)
        _parse_return_code(res, "invert_changeset")

    def create_rebased_changeset(self, base, modified, changeset_their, changeset, conflict):
        func = self.lib.GEODIFF_createRebasedChangeset
        func.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = base.encode('utf-8')
        b_string2 = modified.encode('utf-8')
        b_string3 = changeset_their.encode('utf-8')
        b_string4 = changeset.encode('utf-8')
        b_string5 = conflict.encode('utf-8')

        res = func(b_string1, b_string2, b_string3, b_string4, b_string5)
        _parse_return_code(res, "createRebasedChangeset")

    def rebase(self, base, modified_their, modified, conflict):
        func = self.lib.GEODIFF_rebase
        func.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = base.encode('utf-8')
        b_string2 = modified_their.encode('utf-8')
        b_string3 = modified.encode('utf-8')
        b_string4 = conflict.encode('utf-8')
        res = func(b_string1, b_string2, b_string3, b_string4)
        _parse_return_code(res, "rebase")

    def apply_changeset(self, base, changeset):
        func = self.lib.GEODIFF_applyChangeset
        func.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = base.encode('utf-8')
        b_string2 = changeset.encode('utf-8')

        res = func(b_string1, b_string2)
        _parse_return_code(res, "apply_changeset")

    def list_changes(self, changeset, result):
        func = self.lib.GEODIFF_listChanges
        func.argtypes = [ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = changeset.encode('utf-8')
        b_string2 = result.encode('utf-8')
        res = func(b_string1, b_string2)
        _parse_return_code(res, "list_changes")

    def list_changes_summary(self, changeset, result):
        func = self.lib.GEODIFF_listChangesSummary
        func.argtypes = [ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = changeset.encode('utf-8')
        b_string2 = result.encode('utf-8')
        res = func(b_string1, b_string2)
        _parse_return_code(res, "list_changes_summary")

    def has_changes(self, changeset):
        func = self.lib.GEODIFF_hasChanges
        func.argtypes = [ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = changeset.encode('utf-8')

        nchanges = func(b_string1)
        if nchanges < 0:
            raise GeoDiffLibError("has_changes")
        return nchanges == 1

    def changes_count(self, changeset):
        func = self.lib.GEODIFF_changesCount
        func.argtypes = [ctypes.c_char_p]
        func.restype = ctypes.c_int

        # create byte objects from the strings
        b_string1 = changeset.encode('utf-8')

        nchanges = func(b_string1)
        if nchanges < 0:
            raise GeoDiffLibError("changes_count")
        return nchanges



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# For a fully annotated version of this file and what it does, see
# https://github.com/pypa/sampleproject/blob/master/setup.py

# To upload this file to PyPI you must build it then upload it:
# python setup.py sdist bdist_wheel  # build in 'dist' folder
# python-m twine upload dist/*  # 'twine' must be installed: 'pip install twine'

import io
import re
import os
from setuptools import find_packages
#from setuptools import setup
from skbuild import setup

VERSION = '0.8.0'

setup(
    name="pygeodiff",
    version=VERSION,
    author="Peter Petrik",
    author_email="peter.petrik@lutraconsulting.co.uk",
    description="Python wrapper around GeoDiff library",
    long_description="Python wrapper around GeoDiff library",
    url="https://github.com/lutraconsulting/geodiff",
    packages=["pygeodiff"],
    include_package_data=True,
    keywords=["diff", "gis", "geo", "geopackage", "merge"],
    scripts=[],
    entry_points={"console_scripts": ["pygeodiff=pygeodiff.main:main"]},
    zip_safe=False,
    cmake_args=['-DENABLE_TESTS:BOOL=OFF', '-DENABLE_COVERAGE:BOOL=OFF', '-DBUILD_TOOLS:BOOL=OFF', '-DPYGEODIFFVERSION='+str(VERSION)],
    cmake_source_dir="geodiff",
    cmake_with_sdist=False,
    test_suite="tests.test_project",
    python_requires=">=3.6",
    license="License :: OSI Approved :: MIT License",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)